import re

import babel
import babel.numbers
from babel.numbers import LC_NUMERIC
from communal.enum import StringEnum


class CurrencyCode:
    def __init__(self, code):
        if isinstance(code, CurrencyCode):
            self.code = code
        elif isinstance(code, str):
            self.validate(code)
            self.code = code
        else:
            raise TypeError(
                "First argument given to CurrencyCode constructor should be "
                "either an instance of CurrencyCode or valid three letter "
                "currency code."
            )

    @classmethod
    def validate(self, code, locale=LC_NUMERIC):
        loc = babel.Locale.parse(locale)
        try:
            return loc.currencies[code]
        except KeyError as e:
            raise ValueError(f"'{code}' is not valid currency code.") from e

    @property
    def symbol(self):
        return babel.numbers.get_currency_symbol(
            self.code,
        )

    def __eq__(self, other):
        if isinstance(other, CurrencyCode):
            return self.code == other.code
        elif isinstance(other, str):
            return self.code == other
        else:
            return NotImplemented

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash(self.code)

    def __repr__(self):
        return self.code

    def __str__(self):
        return self.code


CurrencyCode = StringEnum(
    "CurrencyCode",
    [
        (currency, CurrencyCode(currency))
        for currency in babel.Locale(LC_NUMERIC).currencies
    ],
)


class CurrencyParser:
    currency_decimal_symbols = {
        babel.numbers.get_decimal_symbol(loc)
        for loc in babel.localedata.locale_identifiers()
    }
    currency_group_symbols = {
        babel.numbers.get_group_symbol(loc)
        for loc in babel.localedata.locale_identifiers()
    }
    currency_minus_symbols = {
        babel.numbers.get_minus_sign_symbol(loc)
        for loc in babel.localedata.locale_identifiers()
    }
    currency_plus_symbols = {
        babel.numbers.get_plus_sign_symbol(loc)
        for loc in babel.localedata.locale_identifiers()
    }
    currency_symbols = {
        babel.numbers.get_currency_symbol(str(c)): str(c) for c in CurrencyCode
    }

    currency_pattern = "({symbols})|([\\d{numeric}]+)".format(
        symbols="|".join(
            [
                re.escape(s)
                for s in set(currency_symbols.keys()) | set(currency_symbols.values())
            ]
        ),
        numeric="|".join(
            [
                re.escape(s)
                for s in (
                    currency_decimal_symbols
                    | currency_group_symbols
                    | currency_minus_symbols
                    | currency_plus_symbols
                )
            ]
        ),
    )

    group_pattern = re.compile("[{}]".format("".join(currency_group_symbols)))
    numeric_pattern = re.compile("[{}]".format("".join(currency_group_symbols)))

    currency_regex = re.compile(currency_pattern)

    @classmethod
    def extract(cls, s, locale=LC_NUMERIC):
        matches = cls.currency_regex.findall(s)
        currency = None
        value = None
        for c, v in matches:
            if c and not currency:
                currency = c
            if v and not value:
                value = v

            if currency and value:
                break

        loc = babel.Locale.parse(locale)
        value = babel.numbers.parse_decimal(value, locale=loc)

        if currency:
            currency = CurrencyCode(cls.currency_symbols.get(currency, currency))

        return value, currency



from email.utils import parseaddr
from functools import total_ordering

from communal.enum import StringEnum
from sartorial import JSONSchemaFormatted, Serializable


class EmailProvider(StringEnum):
    FASTMAIL = "fastmail"
    GMAIL = "gmail"
    MICROSOFT = "microsoft"
    YAHOO = "yahoo"


@total_ordering
class Email(str, JSONSchemaFormatted, Serializable):
    __schema_format__ = "email"

    TLD_MAP = {
        "con": "com",
    }

    DOMAIN_MISSPELLINGS = {
        "gamil.com": "gmail.com",
        "gogglemail.com": "googlemail.com",
        "googlmail.com": "googlemail.com",
        "goglemail.com": "googlemail.com",
        "hotmial.com": "hotmail.com",
        "homtail.com": "hotmail.com",
        "hoitmail.com": "hotmail.com",
        "homail.com": "hotmail.com",
        "homil.com": "hotmail.com",
        "hotmaill.com": "hotmail.com",
        "yaho.com": "yahoo.com",
        "uahoo.com": "yahoo.com",
        "ayhoo.com": "yahoo.com",
    }

    FASTMAIL_DOMAINS = {"fastmail.com", "messagingengine.com", "fastmail.fm"}
    GMAIL_DOMAINS = {"google.com", "googlemail.com", "gmail.com"}
    MICROSOFT_DOMAINS = {"hotmail.com", "outlook.com", "live.com"}
    YAHOO_DOMAINS = {"yahoodns.net", "yahoo.com", "ymail.com"}

    COMMON_DOMAIN_TLDS = dict(
        [
            domain.split(".", 1)
            for domain in (
                FASTMAIL_DOMAINS | GMAIL_DOMAINS | MICROSOFT_DOMAINS | YAHOO_DOMAINS
            )
        ]
    )

    def __new__(cls, email_address: str, parse_as: EmailProvider = None):
        return str.__new__(cls, email_address)

    def __init__(self, email_address: str, parse_as: EmailProvider = None):
        name, email_address = parseaddr(email_address)
        self.name = name.strip()

        address, domain = email_address.lower().split("@")
        sub_address = None

        if domain in self.COMMON_DOMAIN_TLDS:
            domain = ".".join([domain, self.COMMON_DOMAIN_TLDS[domain]])

        if "." not in domain:
            raise ValueError(f"Invalid email {email_address}: no TLD in {domain}")
        sub_domains, tld = domain.rsplit(".", 1)
        tld = self.TLD_MAP.get(tld, tld)
        domain = ".".join([sub_domains, tld])

        domain = self.DOMAIN_MISSPELLINGS.get(domain, domain)

        domain = ".".join([sub_domains, tld])

        domain = self.DOMAIN_MISSPELLINGS.get(domain, domain)

        self.domain = domain
        self.sub_domains = sub_domains
        self.tld = tld

        # Plus addressing is supported by Microsoft domains and FastMail
        if domain in self.GMAIL_DOMAINS or parse_as == EmailProvider.GMAIL:
            address = address.replace(".", "")
            if "+" in address:
                address, sub_address = address.split("+", 1)
        # Yahoo domain handling of - is like plus addressing
        elif domain in self.YAHOO_DOMAINS or parse_as == EmailProvider.YAHOO:
            if "-" in address:
                address = address.split("-")[0]
        elif domain in self.MICROSOFT_DOMAINS or parse_as == EmailProvider.MICROSOFT:
            if "+" in address:
                address, sub_address = address.split("+", 1)
        # FastMail has domain part username aliasing and plus addressing
        elif domain in self.FASTMAIL_DOMAINS or parse_as == EmailProvider.FASTMAIL:
            domain_segments = domain.split(".")
            if len(domain_segments) > 2:
                address = domain_segments[0]
                domain = ".".join(domain_segments[1:])
            elif "+" in address:
                address, sub_address = address.split("+", 1)

        self.address = address
        self.sub_address = sub_address
        self.normalized = "@".join([address, domain])
        if not sub_address:
            self.full = self.normalized
        else:
            self.full = f"{self.address}+{self.sub_address}@{self.domain}"

    def __str__(self):
        return self.normalized

    def __repr__(self):
        return self.normalized

    def __hash__(self):
        return hash(self.normalized)

    def __lt__(self, other):
        if not other:
            return False
        if not isinstance(other, Email):
            try:
                other = Email(other)
            except (ValueError, TypeError):
                return False
        return self.normalized < other.normalized

    def __eq__(self, other):
        if not other:
            return False
        if not isinstance(other, Email):
            try:
                other = Email(other)
            except (ValueError, TypeError):
                return False
        return self.normalized == other.normalized

    @classmethod
    def parse(cls, email_address: str, parse_as: EmailProvider = None):
        if email_address and isinstance(email_address, str):
            return cls(email_address, parse_as=parse_as)
        else:
            return None

    @classmethod
    def serialize(cls, email_address):
        if isinstance(email_address, Email):
            return str(email_address.normalized)
        return email_address

    @classmethod
    def serialize_full(cls, email_address):
        if isinstance(email_address, Email):
            return str(email_address.full)
        return email_address

    def add_sub_address(self, sub_address):
        if not sub_address:
            return
        self.sub_address = sub_address
        self.full = f"{self.address}+{sub_address}@{self.domain}"

    @classmethod
    def with_sub_address(cls, email_address, sub_address):
        email = Email(email_address)
        if sub_address:
            email.add_sub_address(sub_address)
        return email



import unicodedata
from pathlib import Path
from typing import ForwardRef, Union
from urllib.parse import quote as quote_orig
from urllib.parse import unquote as unquote_orig

import furl
from communal.encoding import safe_decode
from sartorial import JSONSchemaFormatted, Serializable

__all__ = ["URL", "NormalizedURL"]

URL = ForwardRef("URL")


class URL(str, furl.furl, JSONSchemaFormatted, Serializable):
    __schema_format__ = "url"

    DEFAULT_CHARSET = "utf-8"
    DEFAULT_SCHEME = "https"

    SOURCE_QUERY_PARAMS = {
        "src",
        "source",
        "utm_medium",
        "utm_source",
        "utm_campaign",
        "utm_content",
        "utm_term",
    }

    def __new__(cls, url: str = "", is_normalized=False, **kwargs):
        f = furl.furl(url, **kwargs)
        return str.__new__(cls, f)

    def __init__(self, url: str = "", is_normalized=False, **kwargs):
        super().__init__(url, **kwargs)
        self.is_normalized = is_normalized

    def __eq__(self, other):
        if self.is_normalized and not (isinstance(other, URL) and other.is_normalized):
            other = URL.normalize(other)
        elif not self.is_normalized and not isinstance(other, URL):
            other = URL(other)
        return self.url == other.url

    def __hash__(self):
        return hash(self.url)

    def __str__(self):
        return self.url

    def __repr__(self):
        return self.url

    @classmethod
    def validate(cls, value):
        return cls(value)

    @classmethod
    def parse(cls, value):
        if value is not None:
            return cls(value)
        return value

    @classmethod
    def provide_url_scheme(cls, url, default_scheme=DEFAULT_SCHEME):
        """Make sure the URL has a scheme.
        Params:
            url : string : the URL
            default_scheme : string : the default scheme to use, e.g. 'https'
        Returns:
            string : updated url with validated/attached scheme
        """
        has_scheme = "://" in url
        is_universal_scheme = url.startswith("//")
        is_file_path = url == "-" or (url.startswith("/") and not is_universal_scheme)
        if not url or has_scheme or is_file_path:
            return url
        if is_universal_scheme:
            return default_scheme + ":" + url
        return default_scheme + "://" + url

    @classmethod
    def cleanup(cls, url):
        """Clean up the URL, remove shebangs and trailing space/ampersands/question marks.
        Params:
            url : string : the URL
        Returns:
            string : update url
        """
        url = url.replace("#!", "?_escaped_fragment_=")
        url = url.rstrip("&? ")
        return url

    @classmethod
    def normalize_scheme(cls, scheme: str) -> str:
        """Lowercase scheme
        Params:
            scheme : string : url scheme, e.g., 'https'
        Returns:
            string : normalized scheme data.
        """
        return scheme.lower() if scheme is not None else scheme

    @classmethod
    def normalize_host(cls, host: str, charset=DEFAULT_CHARSET) -> str:
        """Normalize the host, lowercase, remove trailing dots, encode IDN domains.
        Params:
            host : string : url host, e.g., 'example.com'
        Returns:
            string : normalized host data.
        """
        if host is None:
            return host
        host = safe_decode(host, charset)
        host = host.lower()
        host = host.strip(".")
        host = host.encode("idna").decode(charset)
        return host

    @classmethod
    def unquote(cls, string, charset="utf-8"):
        """URL decode a string and normalize unicode to NFC
        Params:
            string : string to be unquoted
            charset : string : optional : output encoding
        Returns:
            string : an unquoted and normalized string
        """
        string = unquote_orig(string)
        string = safe_decode(string, charset)
        string = unicodedata.normalize("NFC", string).encode(charset)
        return string

    @classmethod
    def quote(cls, string, safe="/"):
        """URL encode a string, but do not encode the safe characters.
        Params:
            string : string to be quoted
            safe : string of safe characters
        Returns:
            string : quoted string
        """
        string = quote_orig(string, safe)
        return string

    @classmethod
    def normalize_path(cls, path: Union[str, Path], scheme) -> str:
        """Normalize path part of the url. Quote only the necessary parts
        of the URL (e.g., spaces but not parens, etc.)
        Params:
            path : string : url path, e.g., '/path/to/page.html'
            scheme : string : url scheme, e.g., 'https'
        Returns:
            string : normalized path data.
        """
        # Only perform percent-encoding where it is essential.
        # Always use uppercase A-through-F characters when percent-encoding.
        # All portions of the URI must be utf-8 encoded NFC from Unicode strings
        path = cls.quote(cls.unquote(str(path)), "~:/?#[]@!$&'()*+,;=")
        # Prevent dot-segments appearing in non-relative URI paths.
        if scheme is None or scheme in ["", "http", "https", "ftp", "file"]:
            output, part = [], None
            for part in path.split("/"):
                if part == "":
                    if not output:
                        output.append(part)
                elif part == ".":
                    pass
                elif part == "..":
                    if len(output) > 1:
                        output.pop()
                else:
                    output.append(part)
            if part in ["", ".", ".."]:
                output.append("")
            path = "/".join(output)
        # For schemes that define an empty path to be equivalent to a path of "/",
        # use "/".
        if not path and scheme in ["http", "https", "ftp", "file"]:
            path = "/"
        return path

    @classmethod
    def normalize_fragment(cls, fragment: str) -> str:
        """Normalize fragment part of the url, quote all except ~
        Params:
            fragment : string : url fragment, e.g., 'fragment'
        Returns:
            string : normalized fragment data.
        """
        return cls.quote(cls.unquote(str(fragment)), "~")

    @classmethod
    def normalize_query(cls, query, sort_query_params=True):
        """Normalize query part of the url. Optionally sort the query params
        and remove common source code parameters such as utm tags
        Params:
            query : string : url query, e.g., 'param1=val1&param2=val2'
        Returns:
            string : normalized query data.
        """
        if not isinstance(query, furl.Query):
            query = furl.Query(query)
        for k in cls.SOURCE_QUERY_PARAMS:
            query.params.pop(k, None)
        params = [
            (cls.quote(cls.unquote(k)), cls.quote(cls.unquote(v)))
            for k, v in query.params.items()
        ]
        if sort_query_params:
            params = sorted(params)
        return "&".join(f"{k}={v}" for k, v in params)

    @classmethod
    def normalize(
        cls,
        url: str,
        charset=DEFAULT_CHARSET,
        default_scheme=DEFAULT_SCHEME,
        sort_query_params=True,
    ):
        """Normalize the URL to use percent-encoding (only where necessary and always using uppercase A-through-F characters),
        lowercase scheme and host, remove default port numbers, remove dot-segments, remove source code parameters such as utm tags,
        and by default sort the query parameters (though this is optional).
        >>> URL.normalize('https://en.wikipedia.org/wiki/Springfield (toponym)?utm_medium=social&utm_source=foo&utm_campaign=bar&utm_content=blee&utm_term=blah')
        'https://en.wikipedia.org/wiki/Springfield%20(toponym)'
        Params:
            url : string : the URL
            charset : string : the charset to use, e.g. 'utf-8'
            default_scheme : string : default scheme to use, e.g. 'https'
            sort_query_params : bool : whether to sort the query parameters
        Returns:
            URL : normalized url (subclass of str with additional methods and properties)
        """
        if not url:
            return url
        if isinstance(url, URL):
            url = str(url)
        url = cls.provide_url_scheme(url, default_scheme)
        url = cls.cleanup(url)
        url = URL(url)
        url = url.set(
            scheme=cls.normalize_scheme(url.scheme),
            host=cls.normalize_host(url.host, charset),
            query=cls.normalize_query(url.query, sort_query_params),
            fragment=cls.normalize_fragment(url.fragment),
        )
        if not url.username:
            url.username = None

        if not url.password:
            url.password = None

        url = url.set(
            path=cls.normalize_path(url.path, url.scheme),
        )
        url.is_normalized = True
        return url


class NormalizedURL(URL):
    __schema_format__ = "normalized-url"

    def __init__(self, url: str = "", **kwargs):
        u = URL.normalize(url)
        super().__init__(u.url, is_normalized=True, **kwargs)



from decimal import Decimal
from typing import ForwardRef, Union

from babel import Locale
from babel.core import default_locale
from babel.numbers import format_currency, get_decimal_symbol
from money import Money as BaseMoney
from sartorial import JSONSchemaFormatted, Serializable

from atypical.currency import CurrencyCode, CurrencyParser

Money = ForwardRef("Money")


class Money(str, BaseMoney, JSONSchemaFormatted, Serializable):
    __schema_format__ = "money"
    DEFAULT_LOCALE = None

    def __new__(
        cls,
        amount: Union[Decimal, str, int, Money],
        currency: Union[CurrencyCode, str] = CurrencyCode.USD,
        locale: Union[Locale, str] = None,
    ):
        return str.__new__(cls, amount)

    def __init__(
        self,
        amount: Union[Decimal, str, int, Money],
        currency: Union[CurrencyCode, str] = CurrencyCode.USD,
        locale: Union[Locale, str] = None,
    ):
        if isinstance(amount, BaseMoney):
            super().__init__(amount.amount, getattr(amount, "currency", str(currency)))
        elif isinstance(amount, str) and amount:
            value, given_currency = CurrencyParser.extract(amount)
            if given_currency:
                currency = given_currency
            super().__init__(value, str(currency))
        elif not amount:
            super().__init__(0, str(currency))
        else:
            super().__init__(amount, str(currency))
        if locale is None:
            if self.__class__.DEFAULT_LOCALE is None:
                self.__class__.DEFAULT_LOCALE = default_locale("LC_NUMERIC")
            locale = self.__class__.DEFAULT_LOCALE
        self.locale = locale
        self.formatted = format_currency(self.amount, self.currency, locale=self.locale)

    def __str__(self):
        return self.formatted

    def __repr__(self):
        return self.formatted

    def __hash__(self) -> int:
        return BaseMoney.__hash__(self)

    def __lt__(self, other):
        return BaseMoney.__lt__(self, Money(other))

    def __le__(self, other):
        return BaseMoney.__le__(self, Money(other))

    def __eq__(self, other):
        return BaseMoney.__eq__(self, Money(other))

    def __ne__(self, other):
        return BaseMoney.__ne__(self, Money(other))

    def __gt__(self, other):
        return BaseMoney.__gt__(self, Money(other))

    def __ge__(self, other):
        return BaseMoney.__ge__(self, Money(other))

    def __bool__(self):
        return BaseMoney.__bool__(self)

    def __add__(self, other):
        return Money(BaseMoney.__add__(self, Money(other)))

    def __radd__(self, other):
        return Money(BaseMoney.__radd__(self, Money(other)))

    def __sub__(self, other):
        return Money(BaseMoney.__sub__(self, Money(other)))

    def __rsub__(self, other):
        return Money(BaseMoney.__rsub__(self, Money(other)))

    def __mul__(self, other):
        return Money(BaseMoney.__mul__(self, Money(other)))

    def __rmul__(self, other):
        return Money(BaseMoney.__rmul__(self, Money(other)))

    def __truediv__(self, other):
        return Money(BaseMoney.__truediv__(self, Money(other)))

    def __floordiv__(self, other):
        return Money(BaseMoney.__floordiv__(self, Money(other)))

    def __mod__(self, other):
        return Money(BaseMoney.__mod__(self, Money(other)))

    def __divmod__(self, other):
        return Money(BaseMoney.__divmod__(self, Money(other)))

    def __pow__(self, other):
        return Money(BaseMoney.__pow__(self, Money(other)))

    def __neg__(self):
        return Money(BaseMoney.__neg__(self))

    def __pos__(self):
        return Money(BaseMoney.__pos__(self))

    def __abs__(self):
        return Money(BaseMoney.__abs__(self))

    def __int__(self):
        return int(self.amount)

    def __float__(self):
        return float(self.amount)

    def __round__(self, ndigits=None):
        return Money(BaseMoney.__round__(self, ndigits))

    @property
    def friendly(self) -> str:
        symbol = get_decimal_symbol(self.locale)
        return self.formatted.replace(f"{symbol}00", "")

    @property
    def num_cents(self) -> int:
        return int(self.amount * 100)

    def as_cents(self) -> int:
        return self.num_cents

    @classmethod
    def to_cents(cls, amount: Union[Decimal, str, int, Money]) -> int:
        return cls(amount=amount).as_cents()

    @classmethod
    def from_cents(
        cls, amount: int, currency: CurrencyCode = CurrencyCode.USD
    ) -> Money:
        return cls(amount=Decimal(amount) / 100, currency=currency)



import re
from typing import Any, ClassVar, Dict, ForwardRef, List, Union

from communal.nulls import Omitted
from jinja2 import Environment, StrictUndefined
from jinja2 import Template as JinjaTemplate
from jinja2 import meta as jinja2_meta
from sartorial import JSONSchemaFormatted, Serializable
from sartorial.types import JSON_SCHEMA_DEFAULT_TYPES

JSON_TYPE_DEFAULT_SCHEMAS = {v: k for k, v in JSON_SCHEMA_DEFAULT_TYPES.items()}

Templated = ForwardRef("Templated")


class Templated(str, JSONSchemaFormatted, Serializable):
    __schema_format__ = "jinja2"

    env: ClassVar[Environment] = Environment(undefined=StrictUndefined)
    jinja_check_pattern: ClassVar[re.Pattern] = re.compile(r"\{\{.*?\}\}|\{\%.*?\%\}")

    template: str
    compiled: JinjaTemplate
    variables: List[str]
    params: Dict[str, Any]

    def __new__(cls, template: str, **kwargs):
        return str.__new__(cls, template)

    def __init__(self, template: Union[str, Templated], **kwargs):
        if isinstance(template, Templated):
            self.params = template.params
            self.template = template.template
            self.compiled = template.compiled
            self.variables = template.variables
            self.params = template.params or {}
            self.params.update(kwargs)
        elif not isinstance(template, str):
            raise ValueError("Template must be a string")
        else:
            self.template = template
            nodes = self.env.parse(template)
            self.variables = list(jinja2_meta.find_undeclared_variables(nodes))
            self.compiled = self.env.template_class.from_code(
                environment=self.env,
                code=self.env.compile(nodes),
                globals=self.env.make_globals(None),
                uptodate=None,
            )
            self.params = kwargs

    def render(self, **kwargs):
        if self.params is not Omitted:
            kwargs = dict(self.params, **kwargs)
        value = self.compiled.render(**kwargs)
        return value

    @classmethod
    def is_jinja(cls, value: str):
        parsed = cls.env.parse(value)
        return bool(jinja2_meta.find_undeclared_variables(parsed))

    def __repr__(self) -> str:
        return self.template

    def __str__(self) -> str:
        return self.template



from functools import total_ordering

try:
    import phonenumbers
    from phonenumbers.phonenumber import PhoneNumber as BasePhoneNumber
    from phonenumbers.phonenumberutil import NumberParseException
except ImportError:
    phonenumbers = None
    BasePhoneNumber = object
    NumberParseException = Exception

from communal.exceptions import ConfigurationError
from sartorial import JSONSchemaFormatted, Serializable


class PhoneNumberParseException(NumberParseException):
    """
    Catch this as either a PhoneNumberParseException or
    NumberParseException from the phonenumbers library.
    """

    pass


@total_ordering
class PhoneNumber(str, BasePhoneNumber, JSONSchemaFormatted, Serializable):
    """
    Wrapper class for parsing a phone number using the phonenumbers library,
    based on libphonenumber from Google.

    .. phonenumbers library:
       https://github.com/daviddrysdale/python-phonenumbers

    :param raw_number:
        String representation of the phone number.
    :param country:
        Country code of the phone number.
    :param check_region:
        Whether to check the supplied region parameter;
        should always be True for external callers.
        Can be useful for short codes or toll free
    """

    DEFAULT_REGION = "US"
    DEFAULT_FORMAT = "standard"

    __schema_format__ = "phone"

    def __new__(
        cls, raw_number, region: str = DEFAULT_REGION, check_region: bool = True
    ):
        return str.__new__(cls, raw_number)

    def __init__(
        self, raw_number, region: str = DEFAULT_REGION, check_region: bool = True
    ):
        # Bail if phonenumbers is not found.
        if phonenumbers is None:
            raise ConfigurationError("'phonenumbers' is required to use 'PhoneNumber'")

        if isinstance(raw_number, str):
            try:
                self._phone_number = phonenumbers.parse(
                    raw_number, region, _check_region=check_region
                )
            except NumberParseException as e:
                raise PhoneNumberParseException(
                    getattr(e, "error_type", -1), str(e)
                ) from e
        elif isinstance(raw_number, BasePhoneNumber):
            self._phone_number = raw_number
        elif isinstance(raw_number, PhoneNumber):
            self._phone_number = raw_number._phone_number

        region = phonenumbers.region_code_for_country_code(
            self._phone_number.country_code
        )

        super().__init__(
            country_code=self._phone_number.country_code,
            national_number=self._phone_number.national_number,
            extension=self._phone_number.extension,
            italian_leading_zero=self._phone_number.italian_leading_zero,
            raw_input=self._phone_number.raw_input,
            country_code_source=self._phone_number.country_code_source,
            preferred_domestic_carrier_code=(
                self._phone_number.preferred_domestic_carrier_code
            ),
        )
        self.region = region
        self.national = phonenumbers.format_number(
            self._phone_number, phonenumbers.PhoneNumberFormat.NATIONAL
        )
        self.friendly = self.national
        self.international = phonenumbers.format_number(
            self._phone_number, phonenumbers.PhoneNumberFormat.INTERNATIONAL
        )
        self.standard = phonenumbers.format_number(
            self._phone_number, phonenumbers.PhoneNumberFormat.E164
        )

    def __composite_values__(self):
        return self.national, self.region

    def is_valid_number(self):
        return phonenumbers.is_valid_number(self._phone_number)

    @property
    def area_code(self):
        if self.country_code == 1:
            return str(self.national_number)[:3]
        return None

    def __str__(self):
        return self.standard

    def __repr__(self):
        return self.standard

    def __hash__(self):
        return hash(self.standard)

    def __lt__(self, other):
        if not other:
            return False
        if not isinstance(other, PhoneNumber):
            try:
                other = PhoneNumber(str(other))
            except (PhoneNumberParseException, TypeError):
                return False
        return self.standard < other.standard

    def __eq__(self, other):
        if not other:
            return False
        if not isinstance(other, PhoneNumber):
            try:
                other = PhoneNumber(str(other))
            except (PhoneNumberParseException, TypeError):
                return False
        return self.standard == other.standard

    def to_string(self) -> str:
        return self.standard

    @classmethod
    def parse(cls, value, region=DEFAULT_REGION, check_region=True):
        return PhoneNumber(value, region=region, check_region=check_region)

    @classmethod
    def serialize(cls, value, format=DEFAULT_FORMAT, region=DEFAULT_REGION):
        if value:
            if not isinstance(value, PhoneNumber):
                value = PhoneNumber(value, region=region)

            if format == "standard" and value.extension:
                return f"{value.standard};ext={value.extension}"

            return getattr(value, format)

        return value
